==> cmios9-2.0.tar.gz: cmios9-2.0/cmios9_io.c <==
/*
* Copyright (C) 2001, 2002, 2003, 2004  K. M. Indlekofer
*
* M.Indlekofer@fz-juelich.de
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/



#include "cmios9_err.h"
#include "cmios9_io.h"



/* basic I/O */
/* device */
int readonly;



/* basic I/O */

/* general */
/* XXX no cache */
int
readblock(int devdes,u_int blknr,void *buf,u_int bufsiz,u_int startsec,u_int totsize,u_int blksiz)
{
	OFF_T off;
	int bstart,bend,b1start,b1siz,b2start,b2siz,b3start,b3siz;
	u_int dev_blks;
	static char bbuf[DEV_BLOCKSIZE];
	char *bp;
	
	if (buf==NULL){
		err_print("readblock: buf==NULL\n");
		return -1;
	}
	if (blksiz==0){
		return -1;
	}
	if (bufsiz==0){
		return -1;
	}
	if (blknr+bufsiz>totsize){
		err_print("readblock: blknr out of range\n");
		return -1;
	}
	
	dev_blks=DEV_BLOCKSIZE/blksiz; /* number of blocks per device block */
	
	/* parameters for 3 blocks */
	bstart=startsec+blknr;
	bend=bstart+bufsiz;
	b1start=dev_blks*(bstart/dev_blks);
	b2start=dev_blks*((bstart+dev_blks-1)/dev_blks);
	b3start=dev_blks*(bend/dev_blks);
	b1siz=b2start-bstart; /* should be smaller than dev_blks now! */
	if (b1siz>=(int)bufsiz){
		/* just block 1 */
		b1siz=bufsiz;
		b2siz=0;
		b3siz=0;
	}else{
		/* block 1 is not enough */
		b2siz=b3start-b2start; /* should be multiple of dev_blks now! */
		if (b2siz<0){
			/* not enough for block 2 */
			b2siz=0;
		}
		b3siz=bend-b3start; /* should be smaller than dev_blks now! */
	}
	/* XXX off_t to prevent overflow of u_int!!! */
	off=((OFF_T)b1start)*((OFF_T)blksiz); /* lseek offset, multiple of dev_blks */
#ifdef DEBUGIO
	printf("---\n");
	printf("readblock: blknr:   %06x\n",blknr);
	printf("readblock: bufsiz:  %06x\n",bufsiz);
	printf("readblock: bstart:  %06x\n",bstart);
	printf("readblock: bend:    %06x\n",bend);
	printf("readblock: b1start: %06x\n",b1start);
	printf("readblock: b1siz:   %06x\n",b1siz);
	printf("readblock: b2start: %06x\n",b2start);
	printf("readblock: b2siz:   %06x\n",b2siz);
	printf("readblock: b3start: %06x\n",b3start);
	printf("readblock: b3siz:   %06x\n",b3siz);
	printf("---\n");
#endif
	bp=(char *)buf;
	
	if (LSEEK(devdes,off,SEEK_SET)!=off){
#ifdef DEBUGIO
		perror("readblock: lseek 1");
#endif
		return -1;
	}
	
	/* block 1 */
	if (b1siz>0){
		if (read(devdes,bbuf,dev_blks*blksiz)!=(int)(dev_blks*blksiz)){
#ifdef DEBUGIO
			perror("readblock: read 1");
#endif
			return -1;
		}
		/* extract part */
		bcopy(bbuf+(bstart-b1start)*blksiz,bp,b1siz*blksiz);
		bp+=b1siz*blksiz;
	}
	
	/* block 2 */
	/* XXX b2siz*blksiz must not exceed SSIZE_MAX! */
	if (b2siz>(int)(SSIZE_MAX/blksiz)){
		err_print("readblock: bufsiz too large\n");
		return -1;
	}
	if (b2siz>0){
		if (read(devdes,bp,b2siz*blksiz)!=(int)(b2siz*blksiz)){
#ifdef DEBUGIO
			perror("readblock: read 2");
#endif
			return -1;
		}
		bp+=b2siz*blksiz;
	}
	
	/* block 3 */
	if (b3siz>0){
		/* at least b3siz */
		if (read(devdes,bbuf,dev_blks*blksiz)<(int)(b3siz*blksiz)){
#ifdef DEBUGIO
			perror("readblock: read 3");
#endif
			return -1;
		}
		/* extract part */
		bcopy(bbuf,bp,b3siz*blksiz);
		bp+=b3siz*blksiz;
	}
	
	return 0;
}

/* general */
/* XXX no cache */
int
writeblock(int devdes,u_int blknr,void *buf,u_int bufsiz,u_int startsec,u_int totsize,u_int blksiz)
{
	OFF_T off;
	int bstart,bend,b1start,b1siz,b2start,b2siz,b3start,b3siz;
	u_int dev_blks;
	static char bbuf[DEV_BLOCKSIZE];
	char *bp;
	
	if (buf==NULL){
		err_print("writeblock: buf==NULL\n");
		return -1;
	}
	if (blksiz==0){
		return -1;
	}
	if (bufsiz==0){
		return -1;
	}
	if (blknr+bufsiz>totsize){
		err_print("writeblock: blknr out of range\n");
		return -1;
	}

	dev_blks=DEV_BLOCKSIZE/blksiz; /* number of blocks per device block */
	
	/* parameters for 3 blocks */
	bstart=startsec+blknr;
	bend=bstart+bufsiz;
	b1start=dev_blks*(bstart/dev_blks);
	b2start=dev_blks*((bstart+dev_blks-1)/dev_blks);
	b3start=dev_blks*(bend/dev_blks);
	b1siz=b2start-bstart; /* should be smaller than dev_blks now! */
	if (b1siz>=(int)bufsiz){
		/* just block 1 */
		b1siz=bufsiz;
		b2siz=0;
		b3siz=0;
	}else{
		/* block 1 is not enough */
		b2siz=b3start-b2start; /* should be multiple of dev_blks now! */
		if (b2siz<0){
			/* not enough for block 2 */
			b2siz=0;
		}
		b3siz=bend-b3start; /* should be smaller than dev_blks now! */
	}
	/* XXX off_t to prevent overflow of u_int!!! */
	off=((OFF_T)b1start)*((OFF_T)blksiz); /* lseek offset, multiple of dev_blks */
#ifdef DEBUGIO
	printf("---\n");
	if (readonly)
		printf("writeblock: READONLY\n");
	printf("writeblock: blknr:   %06x\n",blknr);
	printf("writeblock: bufsiz:  %06x\n",bufsiz);
	printf("writeblock: bstart:  %06x\n",bstart);
	printf("writeblock: bend:    %06x\n",bend);
	printf("writeblock: b1start: %06x\n",b1start);
	printf("writeblock: b1siz:   %06x\n",b1siz);
	printf("writeblock: b2start: %06x\n",b2start);
	printf("writeblock: b2siz:   %06x\n",b2siz);
	printf("writeblock: b3start: %06x\n",b3start);
	printf("writeblock: b3siz:   %06x\n",b3siz);
	printf("---\n");
#endif
	bp=(char *)buf;
	
	if (LSEEK(devdes,off,SEEK_SET)!=off){
#ifdef DEBUGIO
		perror("writeblock: lseek 1");
#endif
		return -1;
	}
	
	/* block 1 */
	if (b1siz>0){
		/* read block */
		if (read(devdes,bbuf,dev_blks*blksiz)!=(int)(dev_blks*blksiz)){
#ifdef DEBUGIO
			perror("writeblock: read 1");
#endif
			return -1;
		}
		/* undo read offset */
		if (LSEEK(devdes,-(OFF_T)(dev_blks*blksiz),SEEK_CUR)<0){
#ifdef DEBUGIO
			perror("writeblock: lseek undo 1");
#endif
			return -1;
		}
		/* modify */
		bcopy(bp,bbuf+(bstart-b1start)*blksiz,b1siz*blksiz);
		/* write block */
#ifndef READONLY
		if (!readonly){
			if (write(devdes,bbuf,dev_blks*blksiz)!=(int)(dev_blks*blksiz)){
#ifdef DEBUGIO
				perror("writeblock: write 1");
#endif
				return -1;
			}
		}
#endif
		bp+=b1siz*blksiz;
	}
	
	/* block 2 */
	/* XXX b2siz*blksiz must not exceed SSIZE_MAX! */
	if (b2siz>(int)(SSIZE_MAX/blksiz)){
		err_print("writeblock: bufsiz too large\n");
		return -1;
	}
	if (b2siz>0){
#ifndef READONLY
		if (!readonly){
			if (write(devdes,bp,b2siz*blksiz)!=(int)(b2siz*blksiz)){
#ifdef DEBUGIO
				perror("writeblock: write 2");
#endif
				return -1;
			}
		}
#endif
		bp+=b2siz*blksiz;
	}
	
	/* block 3 */
	/* XXX assume that we can access full dev_blks at the end */
	if (b3siz>0){
		int roff;

		/* read block */
		/* at least b3siz */
		if ((roff=read(devdes,bbuf,dev_blks*blksiz))<(int)(b3siz*blksiz)){
#ifdef DEBUGIO
			perror("writeblock: read 3");
#endif
			return -1;
		}
		/* undo read offset */
		if (LSEEK(devdes,-(OFF_T)(roff),SEEK_CUR)<0){ /* roff!!! */
#ifdef DEBUGIO
			perror("writeblock: lseek undo 3");
#endif
			return -1;
		}
		/* modify */
		bcopy(bp,bbuf,b3siz*blksiz);
		/* write block */
#ifndef READONLY
		if (!readonly){
			/* at least b3siz */
			if (write(devdes,bbuf,dev_blks*blksiz)<(int)(b3siz*blksiz)){
#ifdef DEBUGIO
				perror("writeblock: write 3");
#endif
				return -1;
			}
		}
#endif
		bp+=b3siz*blksiz;
	}
	
	return 0;
}

int
opendevice(char *devpath)
{
	int devdes;
	
	if (devpath==NULL){
		return -1;
	}

	/* open device/image file */
	/* O_BINARY important for Windows-based systems (esp. cygwin and VC++!) */
	if ((devdes=open(devpath,(readonly?O_RDONLY:O_RDWR)|O_BINARY,0))<0){
		perror("opendevice");
		return -1;
	}

	return devdes;
}




/* EOF */

==> cmios9-2.1.tar.gz: cmios9-2.1/cmios9_io.c <==
/*
* Copyright (C) 2001, 2002, 2003, 2004, 2005  K. M. Indlekofer
*
* M.Indlekofer@fz-juelich.de
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/



#include "cmios9_err.h"
#include "cmios9_io.h"



/* basic I/O */
/* device */
int readonly;



/* basic I/O */

/* general */
/* XXX no cache */
int
readblock(int devdes,u_int blknr,void *buf,u_int bufsiz,u_int startsec,u_int totsize,u_int blksiz)
{
	OFF_T off;
	int bstart,bend,b1start,b1siz,b2start,b2siz,b3start,b3siz;
	u_int dev_blks;
	static char bbuf[DEV_BLOCKSIZE];
	char *bp;
	
	if (buf==NULL){
		err_print("readblock: buf==NULL\n");
		return -1;
	}
	if (blksiz==0){
		return -1;
	}
	if (bufsiz==0){
		return -1;
	}
	if (blksiz==0){
		return -1;
	}
	if (blknr+bufsiz>totsize){
		err_print("readblock: blknr out of range\n");
		return -1;
	}
	
	dev_blks=DEV_BLOCKSIZE/blksiz; /* number of blocks per device block */
	if ((dev_blks==0)||(DEV_BLOCKSIZE%blksiz!=0)){
		dev_blks=1;
	}
	
	/* parameters for 3 chunks */
	bstart=startsec+blknr;
	bend=bstart+bufsiz;
	b1start=dev_blks*(bstart/dev_blks);
	b2start=dev_blks*((bstart+dev_blks-1)/dev_blks);
	b3start=dev_blks*(bend/dev_blks);
	b1siz=b2start-bstart; /* should be smaller than dev_blks now! */
	if (b1siz>=(int)bufsiz){
		/* just chunk 1 */
		b1siz=bufsiz;
		b2siz=0;
		b3siz=0;
	}else{
		/* chunk 1 is not enough */
		b2siz=b3start-b2start; /* should be multiple of dev_blks now! */
		if (b2siz<0){
			/* not enough for chunk 2 */
			b2siz=0;
		}
		b3siz=bend-b3start; /* should be smaller than dev_blks now! */
	}
	/* XXX off_t to prevent overflow of u_int!!! */
	off=((OFF_T)b1start)*((OFF_T)blksiz); /* lseek offset, multiple of dev_blks */
#ifdef DEBUGIO
	printf("---\n");
	printf("readblock: blknr:   %06x\n",blknr);
	printf("readblock: bufsiz:  %06x\n",bufsiz);
	printf("readblock: bstart:  %06x\n",bstart);
	printf("readblock: bend:    %06x\n",bend);
	printf("readblock: b1start: %06x\n",b1start);
	printf("readblock: b1siz:   %06x\n",b1siz);
	printf("readblock: b2start: %06x\n",b2start);
	printf("readblock: b2siz:   %06x\n",b2siz);
	printf("readblock: b3start: %06x\n",b3start);
	printf("readblock: b3siz:   %06x\n",b3siz);
	printf("---\n");
#endif
	bp=(char *)buf;
	
	if (LSEEK(devdes,off,SEEK_SET)!=off){
#ifdef DEBUGIO
		perror("readblock: lseek 1");
#endif
		return -1;
	}
	
	/* chnuk 1 */
	if (b1siz>0){
		if (read(devdes,bbuf,dev_blks*blksiz)!=(int)(dev_blks*blksiz)){
#ifdef DEBUGIO
			perror("readblock: read 1");
#endif
			return -1;
		}
		/* extract part */
		bcopy(bbuf+(bstart-b1start)*blksiz,bp,b1siz*blksiz);
		bp+=b1siz*blksiz;
	}
	
	/* chunk 2 */
	/* XXX b2siz*blksiz must not exceed SSIZE_MAX! */
	if (b2siz>(int)(SSIZE_MAX/blksiz)){
		err_print("readblock: bufsiz too large\n");
		return -1;
	}
	if (b2siz>0){
		if (read(devdes,bp,b2siz*blksiz)!=(int)(b2siz*blksiz)){
#ifdef DEBUGIO
			perror("readblock: read 2");
#endif
			return -1;
		}
		bp+=b2siz*blksiz;
	}
	
	/* chunk 3 */
	if (b3siz>0){
		/* at least b3siz */
		if (read(devdes,bbuf,dev_blks*blksiz)<(int)(b3siz*blksiz)){
#ifdef DEBUGIO
			perror("readblock: read 3");
#endif
			return -1;
		}
		/* extract part */
		bcopy(bbuf,bp,b3siz*blksiz);
		bp+=b3siz*blksiz;
	}
	
	return 0;
}

/* general */
/* XXX no cache */
int
writeblock(int devdes,u_int blknr,void *buf,u_int bufsiz,u_int startsec,u_int totsize,u_int blksiz)
{
	OFF_T off;
	int bstart,bend,b1start,b1siz,b2start,b2siz,b3start,b3siz;
	u_int dev_blks;
	static char bbuf[DEV_BLOCKSIZE];
	char *bp;
	
	if (buf==NULL){
		err_print("writeblock: buf==NULL\n");
		return -1;
	}
	if (blksiz==0){
		return -1;
	}
	if (bufsiz==0){
		return -1;
	}
	if (blksiz==0){
		return -1;
	}
	if (blknr+bufsiz>totsize){
		err_print("writeblock: blknr out of range\n");
		return -1;
	}

	dev_blks=DEV_BLOCKSIZE/blksiz; /* number of blocks per device block */
	if ((dev_blks==0)||(DEV_BLOCKSIZE%blksiz!=0)){
		dev_blks=1;
	}
	
	/* parameters for 3 chunks */
	bstart=startsec+blknr;
	bend=bstart+bufsiz;
	b1start=dev_blks*(bstart/dev_blks);
	b2start=dev_blks*((bstart+dev_blks-1)/dev_blks);
	b3start=dev_blks*(bend/dev_blks);
	b1siz=b2start-bstart; /* should be smaller than dev_blks now! */
	if (b1siz>=(int)bufsiz){
		/* just chunk 1 */
		b1siz=bufsiz;
		b2siz=0;
		b3siz=0;
	}else{
		/* chunk 1 is not enough */
		b2siz=b3start-b2start; /* should be multiple of dev_blks now! */
		if (b2siz<0){
			/* not enough for chunk 2 */
			b2siz=0;
		}
		b3siz=bend-b3start; /* should be smaller than dev_blks now! */
	}
	/* XXX off_t to prevent overflow of u_int!!! */
	off=((OFF_T)b1start)*((OFF_T)blksiz); /* lseek offset, multiple of dev_blks */
#ifdef DEBUGIO
	printf("---\n");
	if (readonly)
		printf("writeblock: READONLY\n");
	printf("writeblock: blknr:   %06x\n",blknr);
	printf("writeblock: bufsiz:  %06x\n",bufsiz);
	printf("writeblock: bstart:  %06x\n",bstart);
	printf("writeblock: bend:    %06x\n",bend);
	printf("writeblock: b1start: %06x\n",b1start);
	printf("writeblock: b1siz:   %06x\n",b1siz);
	printf("writeblock: b2start: %06x\n",b2start);
	printf("writeblock: b2siz:   %06x\n",b2siz);
	printf("writeblock: b3start: %06x\n",b3start);
	printf("writeblock: b3siz:   %06x\n",b3siz);
	printf("---\n");
#endif
	bp=(char *)buf;
	
	if (LSEEK(devdes,off,SEEK_SET)!=off){
#ifdef DEBUGIO
		perror("writeblock: lseek 1");
#endif
		return -1;
	}
	
	/* chunk 1 */
	if (b1siz>0){
		/* read chunk */
		if (read(devdes,bbuf,dev_blks*blksiz)!=(int)(dev_blks*blksiz)){
#ifdef DEBUGIO
			perror("writeblock: read 1");
#endif
			return -1;
		}
		/* undo read offset */
		if (LSEEK(devdes,-(OFF_T)(dev_blks*blksiz),SEEK_CUR)<0){
#ifdef DEBUGIO
			perror("writeblock: lseek undo 1");
#endif
			return -1;
		}
		/* modify */
		bcopy(bp,bbuf+(bstart-b1start)*blksiz,b1siz*blksiz);
		/* write chunk */
#ifndef READONLY
		if (!readonly){
			if (write(devdes,bbuf,dev_blks*blksiz)!=(int)(dev_blks*blksiz)){
#ifdef DEBUGIO
				perror("writeblock: write 1");
#endif
				return -1;
			}
		}
#endif
		bp+=b1siz*blksiz;
	}
	
	/* chunk 2 */
	/* XXX b2siz*blksiz must not exceed SSIZE_MAX! */
	if (b2siz>(int)(SSIZE_MAX/blksiz)){
		err_print("writeblock: bufsiz too large\n");
		return -1;
	}
	if (b2siz>0){
#ifndef READONLY
		if (!readonly){
			if (write(devdes,bp,b2siz*blksiz)!=(int)(b2siz*blksiz)){
#ifdef DEBUGIO
				perror("writeblock: write 2");
#endif
				return -1;
			}
		}
#endif
		bp+=b2siz*blksiz;
	}
	
	/* chunk 3 */
	/* XXX assume that we can access full dev_blks at the end */
	if (b3siz>0){
		int roff;

		/* read chunk */
		/* at least b3siz */
		if ((roff=read(devdes,bbuf,dev_blks*blksiz))<(int)(b3siz*blksiz)){
#ifdef DEBUGIO
			perror("writeblock: read 3");
#endif
			return -1;
		}
		/* undo read offset */
		if (LSEEK(devdes,-(OFF_T)(roff),SEEK_CUR)<0){ /* roff!!! */
#ifdef DEBUGIO
			perror("writeblock: lseek undo 3");
#endif
			return -1;
		}
		/* modify */
		bcopy(bp,bbuf,b3siz*blksiz);
		/* write chunk */
#ifndef READONLY
		if (!readonly){
			/* at least b3siz */
			if (write(devdes,bbuf,dev_blks*blksiz)<(int)(b3siz*blksiz)){
#ifdef DEBUGIO
				perror("writeblock: write 3");
#endif
				return -1;
			}
		}
#endif
		bp+=b3siz*blksiz;
	}
	
	return 0;
}

int
opendevice(char *devpath)
{
	int devdes;
	
	if (devpath==NULL){
		return -1;
	}

	/* open device/image file */
	/* O_BINARY important for Windows-based systems (esp. cygwin and VC++!) */
	if ((devdes=open(devpath,(readonly?O_RDONLY:O_RDWR)|O_BINARY,0))<0){
		perror("opendevice");
		return -1;
	}

	return devdes;
}




/* EOF */
